# Copyright David Leonard, 1999. The University of Queensland, Australia.
# $Id: Demo.py,v 1.1.1.1 1999/10/05 23:58:37 leonard Exp $
import lalr_core, sys, traceback

from lalr_core import LeftAssoc

#-- first, create a new grammar object
g = lalr_core.new_parser()

#-- terminals must be added first with add_terminal(sym[,name[,prec[,assoc]]]).
#   The first argument is an arbitrary, unique object identifying the
#   terminal. The second argument is the string name of the terminal.
#   If the 2nd is not given, PyObject_Str() is called on the first arg.
#   The third argument is a precedence level.
#   The fourth arg is the associativity, but it is currently unused.

eof = "eof";	g.add_terminal(eof)
n = "n";	g.add_terminal(n)
plus = "+";	g.add_terminal(plus,	"+", 1, LeftAssoc)
minus = "-";	g.add_terminal(minus,	"-", 2, LeftAssoc)
star = "*";	g.add_terminal(star,	"*", 3, LeftAssoc)
stroke = "/";	g.add_terminal(stroke,	"/", 4, LeftAssoc)
left = "(";	g.add_terminal(left)
right = ")";	g.add_terminal(right)

#-- next, non-terminals are added with add_non_terminal(sym[,name])
G = "G";	g.add_nonterminal(G, "G")
E = "E";	g.add_nonterminal(E, "E")

#-- next, productions are added with add_production().
#   The first argument is the left hand side of the production,
#   The second is the sequence of symbols that it derives.
#   The third argument is a function to call apon reduction, or defaults
#   to the identity function if None.
#   The 3rd arg function is invoked with a tuple consisting of the return
#   results of inner reductions, or terminal tuples returned from the
#   lexer.

# precedence-free evaluation
#T = "T";	g.add_nonterminal(T, "T")
#F = "F";	g.add_nonterminal(F, "F")
#g.add_production(G, [E])
#g.add_production(E, [E, plus, T], 	lambda e,_,t: e+t)
#g.add_production(E, [E, minus, T], 	lambda e,_,t: e-t)
#g.add_production(E, [T])
#g.add_production(T, [T, star, F], 	lambda t,_,f: t*f)
#g.add_production(T, [T, stroke, F], 	lambda t,_,f: t/f)
#g.add_production(T, [F])
#g.add_production(F, [left, E, right], 	lambda _,e,_: e)
#g.add_production(F, [n])

#-- calculator using precedence rules
g.add_production(G, [E])
g.add_production(E, [E, plus, E], 	lambda a,_,b: a+b)
g.add_production(E, [E, minus, E], 	lambda a,_,b: a-b)
g.add_production(E, [E, star, E], 	lambda a,_,b: a*b)
g.add_production(E, [E, stroke, E], 	lambda a,_,b: a/b)
g.add_production(E, [left, E, right], 	lambda _,e,_: e)
g.add_production(E, [n])

g.set_goal_eof(G, eof)

#
# hand-written lexer
#
nextchar = None
def getchar():
	global nextchar
	nxt = nextchar
	nextchar = sys.stdin.read(1)
	return nxt

def nexttoken():
    global nextchar
    if nextchar is None: 
	getchar()
    while 1:
	if nextchar == '': print "(eof)"; raise SystemExit
	if nextchar == '\n' or nextchar == '\r': nextchar = None; return eof,
	if nextchar == '*': getchar(); return star,
	if nextchar == '/': getchar(); return stroke,
	if nextchar == '+': getchar(); return plus,
	if nextchar == '-': getchar(); return minus,
	if nextchar == '(': getchar(); return left,
	if nextchar == ')': getchar(); return right,
	if nextchar in [' ', '\t', '\n', '\r']: getchar(); continue
	if '0' <= nextchar <= '9': 
	    value = 0.0
	    while '0' <= nextchar <= '9':
		    value = 10 * value + (ord(getchar())-ord('0'))
	    if nextchar == '.':
		getchar()
		frac = 0
		mult = 1.0
		while '0' <= nextchar <= '9':
		    mult = mult / 10
		    frac = 10 * frac + (ord(getchar())-ord('0'))
		value = value + mult * frac
	    return n, value
	raise lalr_core.ParseError, "unexpected character `%s'" % nextchar

#-- Finally, g is ready to parse.
#
#   If this is the first time that parse() has been called after
#   a modification to the productions/terminals etc, a DFA will
#   be constructed internally.
#
#   The argument is a function that is repeatedly called for the
#   next token. The parser expects a non-empty tuple from each
#   call. The first element of the tuple must be one of the terminal
#   objects added to the grammar earlier. Exceptions are propagated to
#   the caller.
#   The return value is the same as that returned by the goal symbol's 
#   reduction function.

while 1:
    try:
	print "Enter expression: ",
	res = g.parse(nexttoken)
	print "> ", res
    except lalr_core.ParseError:
	traceback.print_exc()

