/*	$Id: lalr.c,v 1.1.1.1 1999/10/05 23:58:37 leonard Exp $	*/
/*
 * Copyright 1999, David Leonard. The University of Queensland, Australia.
 *
 * DISCLAIMER: ALL MODERN SOFTWARE AND COMPUTER PROGRAMS ARE SO
 * CRAP THAT THEIR AUTHORS KNOW THAT THEY WILL INVARIABLY FALL SHORT
 * OF ANY IMPLIED OR EXPRESS FUNCTIONALITY AND/OR RELIABILITY. AND
 * BECAUSE OF THIS, YOU WILL SEE LARGE FULLY-CAPITALISED DISCLAIMERS
 * AT THE TOP OF SOURCE CODE, PROCLAIMING THAT THE LAZY BASTARD
 * THAT WROTE THE FOLLOWING CODE IS SO AFRAID OF LITIGATION THAT
 * HE/SHE NEEDS TO MAKE A LAST DITCH EFFORT TO INSULATE THEMSELVES
 * FROM LEGAL ACTION BY SAYING SOMETHING LIKE THE FOLLOWING:
 * IF MY SOFTWARE INCURS ANY DAMAGE, I AM NOT RESPONSIBLE, EVEN
 * THOUGH I AM ALMOST DIRECTLY ITS CAUSE. NO GUARANTEES OR WARRANTIES
 * MAY BE MADE ABOUT THIS SOFTWARE'S UTILITY OR RELIABILITY, EVEN
 * IF YOU HAVE PAID MEGABUCKS FOR IT. FOR EXAMPLE, REALLY BIG GRAMMARS WILL
 * OVERFLOW SOME POORLY CHOSEN LIMITS IN THE CODE BELOW, BUT YOU WON'T 
 * KNOWS THIS UNTIL YOUR CRITICAL SYSTEM FAILS CATASTROPHICALLY. HEH, HEH.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "symbol.h"
#include "lalr.h"

/*
 * LALR(1) parser generator based on English algorithms in chapter 7 of
 * T. Pittman and J. Peters. `The Art of Compiler Design: Theory and Practice'
 * Prentice-Hall, 1992.
 * According to the text, LALR(k) parsers were first described by Korenjak
 * in 1969! That was 30 years ago! Cool!
 */

/*
 * Group productions A->w by their left hand side so that finding
 * productions is faster. g->g_derive[A] will point to a list of
 * productions.
 */
void
derive_productions(g)
	grammar_t *g;
{
	symbol_t s;
	production_t *p;

	for (s = 0; s < MAXSYMS; s++)
		g->g_derive[s] = NULL;
	for (p = g->g_prod; p; p = p->p_next_prod) {
		p->p_next_derive = g->g_derive[p->p_lhs];
		g->g_derive[p->p_lhs] = p;
	}
}

/*
 * Compute what the literature calles the 'First' of a sentential substring.
 * This is the set of terminal symbols that can possibly start a string
 * that the input string can derive. If the input string can also derive the
 * empty string, (epsilon) then true is returned. 
 */
int
derive_first(g, len, rhs, f)
	grammar_t *g;
	int len;
	const symbol_t *rhs;
	symset_t *f;
{
	production_t *p;
	symbol_t s;
	int i;
	int haseps;

	for (i = 0; i < len; i++) {
		s = rhs[i];
		/* A terminal means we don't need to continue. */
		if (IS_TERMINAL(g, s)) {
			symsetadd(f, s);
			return 0;
		} else {
			/* See if we have cached the nonterminal's first. */
			if (!symsetin(&g->g_first[s], s)) {
				/* Use the cached symset. */
				symsetunion(f, &g->g_first[s]);
				if (!symsetin(f, EPSILON))
					return 0;
				symsetdel(f, EPSILON);
			} else if (symsetin(f, s)) {
				/* Avoid unbound recursion. */
				return 0;
			} else {
				/* Need to find first of all derivations. */
				symsetadd(f, s);
				haseps = 0;
				for (p = g->g_derive[s]; p; p = 
				    p->p_next_derive) 
					haseps = haseps ||
					    derive_first(g, p->p_nrhs, 
					    p->p_rhs, f);
				symsetdel(f, s);
				if (!haseps)
					return 0;
				/* Continue on if empty is a possibility. */
			}
		}
	}

	/* If we got this far, then epsilon is a possible first */
	return 1;
}

/*
 * Derive and cache the 'first' sets for all single-symbol substrings
 * consisting of the non-terminals. This is the only function that
 * builds the cache, and should speed up the previous function greatly
 * (as it won't ever need to recurse again).
 */
void
derive_firsts(g)
	grammar_t *g;
{
	symbol_t s;
	symset_t ss;

	for (s = g->g_terminals; s < g->g_nonterminals; s++) {
		symsetempty(&g->g_first[s]);
		symsetadd(&g->g_first[s], s);
	}
	for (s = g->g_terminals; s < g->g_nonterminals; s++) {
		symsetempty(&ss);
		derive_first(g, 1, &s, &ss);
		memcpy(&g->g_first[s], &ss, sizeof ss);
	}
}

/*
 * Perform closure over the items in a state with kernel items.
 * This means that for every item A->x.By;z in the kernel, 
 * add items B->.w;z for all productions B->w. That is, expand
 * non-terminals that follow the 'dot'.
 */
void
closure(g, s)
	grammar_t *g;
	state_t *s;
{
	item_t *it, *nit, *iend;
	production_t *p;
	symbol_t nt, follow;
	symset_t nfollow;

	assert(s->s_item != NULL);

	/* Walk to the end of the item list so we can append easily. */
	for (iend = s->s_item; iend->i_next; iend = iend->i_next)
		;

	/* Perform closure over each non-reducing items. */
	for (it = s->s_item; it; it = it->i_next) {
		if (!IS_REDUCE(it)) {
			/* If we can transit a nonterminal... */
			nt = it->i_prod->p_rhs[it->i_dot];
			if (IS_NONTERMINAL(g, nt)) {
			    /* Determine possible symbols after the nt. */
			    symsetempty(&nfollow);
			    if (derive_first(g, 
				it->i_prod->p_nrhs - it->i_dot - 1,
				it->i_prod->p_rhs + it->i_dot + 1, &nfollow))
				    /* epsilon means we use this follow */
				    symsetadd(&nfollow, it->i_follow);
			    /* Add all new derivations with the new follows. */
			    SYMSET_FOREACH(follow, &nfollow)
			    for (p = g->g_derive[nt]; p; p = p->p_next_derive) {
				for (nit = s->s_item; nit; nit = nit->i_next)
				    if (nit->i_prod == p && 
					nit->i_dot == 0 &&
					nit->i_follow == follow)
					break;
				if (nit == NULL) {
				    nit = new_item(p, 0, follow);
				    /* Add to end of item list */
				    assert(iend->i_next == NULL);
				    nit->i_next = NULL;
				    iend->i_next = nit;
				    iend = nit;
				}
			    }
			}
		}
	}
}


void
conflict(g, cls, winner, loser)
	grammar_t *g;
	char *cls;
	item_t *winner, *loser;
{
	fprintf(stderr, "%s conflict:\n\t", cls);
	print_item(g, winner);
	fprintf(stderr, "\t<- winner\n\t");
	print_item(g, loser);
	fprintf(stderr, "\n");
}

/*
 * Add the outgoing shift/reduce transitions to a state, based on
 * all of its items. This is done by grouping the shiftable symbols
 * or reducable productions together. When multiple actions in the
 * one state can occur, their lookahead sets are used. (Conflicts are
 * detected when the lookahead sets intersect, and -1 is returned.)
 * Also, this function generates new kernel-only states when needed.
 */
void
add_transitions(g, st)
	grammar_t *g;
	state_t *st;
{
	symbol_t s;
	state_t *dest;
	item_t *shift[MAXSYMS], *reduce[MAXSYMS];
	item_t *kern[MAXSYMS];
	item_t *it, *nit;
	int redprec, shiftprec, old_redprec;

	/* Reset actions. */
	for (s = 0; s < g->g_nonterminals; s++) {
		kern[s] = NULL;			/* collective shifts */
		shift[s] = NULL;		/* best shift item */
		reduce[s] = NULL;		/* best reduce item */
		st->s_reduce[s] = NULL;		/* reducing production */
		st->s_shift[s] = NULL;		/* new state after shift */
	}

	/* Look for reductions. */
	for (it = st->s_item; it; it = it->i_next)
	    if (IS_REDUCE(it)) {
		s = it->i_follow;
		redprec = it->i_prod->p_prec;
		if (reduce[s]) {
			old_redprec = reduce[s]->i_prod->p_prec;
			if (old_redprec > redprec) {
			    /* conflict(g, "reduce-reduce", reduce[s], it); */
			    continue;
			}

			if (old_redprec < redprec) {
			    /* conflict(g, "reduce-reduce", it, reduce[s]); */
			}

			if (old_redprec == redprec) {
				g->g_rrconflicts++;
				/* arbitrary choice of reduction */
				conflict(g, "reduce-reduce UNRESOLVED", 
				    it, reduce[s]);
			}
		}
		reduce[s] = it;
		st->s_reduce[s] = it->i_prod;
	    }

	/* Look for shifts. */
	for (it = st->s_item; it; it = it->i_next)
	    if (!IS_REDUCE(it)) {
		s = it->i_prod->p_rhs[it->i_dot];
		shiftprec = g->g_symprec[s];
		if (reduce[s]) {
			redprec = reduce[s]->i_prod->p_prec;
			if (redprec > shiftprec) {
			    /* conflict(g, "shift-reduce", reduce[s], it); */
			    continue;
			}

			if (redprec < shiftprec) {
			    /* conflict(g, "shift-reduce", it, reduce[s]); */
			    reduce[s] = NULL;
			}

			if (redprec == shiftprec) {
			    switch (g->g_symassoc[s]) {
			    case NONASSOC:
				g->g_srconflicts++;
				/* arbitrarily allow all shifts to win */
				conflict(g, "shift-reduce UNRESOLVED", 
					it, reduce[s]);
				break;
			    case LEFTASSOC:
				/* reduce left hand side before shifting */
				continue;
			    case RIGHTASSOC:
				/* shift right hand side before shifting */
				break;
			    }
			}
		}
		shift[s] = it;

		/* We can shift this symbol; add shifted item to kernel */
		nit = new_item(it->i_prod, it->i_dot + 1, it->i_follow);
		nit->i_next = kern[s];
		kern[s] = nit;
	    }

	/* Turn all the kernels into new states. */
	for (s = 0; s < g->g_nonterminals; s++) {
		if (kern[s] == NULL)
			continue;
		dest = find_state(g, kern[s]);
		if (dest != NULL)
			free_item_list(kern[s]);
		else {
			dest = new_state(g, kern[s]);
			if (dest == NULL)
				return;
			dest->s_next_new = g->g_newstates;
			g->g_newstates = dest;
		}
		st->s_shift[s] = dest;
	}
}

/*
 * LR(1): Look for another state which has exactly the same kernel
 * as the list of items passed in.
 * LALR(1): Look for another state which has pretty much the same
 * kernel as the list of items passed in - but ignore the lookahead sets.
 * If we find one that is pretty close, augment the lookahead sets.
 * Because of this augmentation, we need to put the state back on the
 * grammar's "new state list" so that the augmentation gets properly
 * propagated.
 */
state_t *
find_state(g, its)
	grammar_t *g;
	item_t *its;
{
	item_t *it, *ic, *nit;
	state_t *s;
	int i;
	int itslen;
	int altered;

	/* Count the length of the its list */
	for (itslen = 0, it = its; it; it = it->i_next, itslen++)
		;

	for (s = g->g_state; s; s = s->s_next) {
	    /* Check that each of our items is in the state's kernel list */
	    for (it = its; it; it = it->i_next) {
	        for (i = 0, ic = s->s_item; i < s->s_nkitem; i++, 
		     ic = ic->i_next) {
		    if (ic->i_prod   == it->i_prod &&
			ic->i_dot    == it->i_dot)
			break;
		}
	        if (i == s->s_nkitem)
			break;	/* missed one */
	    }
	    if (!it)
		break;	/* all matched */
	}

	if (s) {
	    /* Okay, we now have a state that may need to be augmented. */
	    altered = 0;
	    for (it = its; it; it = it->i_next) {
	        for (i = 0, ic = s->s_item; i < s->s_nkitem; i++, 
		     ic = ic->i_next) {
		    if (ic->i_prod   == it->i_prod &&
			ic->i_dot    == it->i_dot &&
			ic->i_follow == it->i_follow)
			break;
		}
		if (i == s->s_nkitem) {
			/* not found exactly, so must merge new bits */
			nit = new_item(it->i_prod, it->i_dot, it->i_follow);
#ifndef NDEBUG
			fprintf(stderr, "Adding item to state %d: ", s->s_id);
			print_item(g, nit);
			fprintf(stderr, "\n");
#endif
			/* Insert nit into s */
			nit->i_next = s->s_item;
			s->s_item = nit;
			s->s_nkitem++;
			altered = 1;
		}
	    }
	    if (altered && !IS_NEW_STATE(s)) {
		s->s_next_new = g->g_newstates;
		g->g_newstates = s;
	    }
	}

#ifndef NDEBUG
	if (s == NULL) 
	    fprintf(stderr, "No match.\n");
	else {
	    fprintf(stderr, "Matched\n");
	    for (it = its; it; it = it->i_next) {
		fprintf(stderr, "\t"); 
		print_item(g, it); 
		fprintf(stderr, "\n");
	    }
	    fprintf(stderr, " with kernel of state %d:\n", s->s_id);
	    for (i=0,it=s->s_item; i<s->s_nkitem; i++,it=it->i_next) {
		fprintf(stderr, "\t");
		print_item(g, it);
		fprintf(stderr, "\n");
	    }
	}
#endif
	return s;
}

static int next_state_id = 0;

/*
 * Create a new state in the grammar. It is initially deemed to be not-new.
 */
state_t *
new_state(g, kern)
	grammar_t *g;
	item_t *kern;
{
	state_t *st;
	item_t *it;

	st = (state_t *)malloc(sizeof *st);
	if (st == NULL) {
		nocore(g);
		return NULL;
	}

	st->s_id = next_state_id++;
	st->s_next = NULL;
	if (g->g_laststate)
	    g->g_laststate->s_next = st;
	else
	    g->g_state = st;
	g->g_laststate = st;
	st->s_next_new = st;	/* point to self if not new */
	st->s_item = kern;

	st->s_nkitem = 0;
	for (it = st->s_item; it; it = it->i_next)
		st->s_nkitem++;

#ifndef NDEBUG
	{ symbol_t s;
	for (s = 0; s < MAXSYMS; s++) {
		st->s_shift[s] = NULL;
		st->s_reduce[s] = NULL;
	}
	fprintf(stderr, "New state:\n");
	print_state(g, st);
	}
#endif

	return st;
}

/* Free a state. */
void
free_state(st)
	state_t *st;
{
	free_item_list(st->s_item);
	free(st);
}

static item_t *item_free_list = NULL;

/* Create a new item. */
item_t *
new_item(p, dot, follow) 
	production_t *p;
	int dot;
	symbol_t follow;
{
	item_t *it;

	/* Do our own memory allocation for a bit more speed. */
	if (item_free_list == NULL) {
		static int allocsz = 128;
		int i;

		it = (item_t *)malloc(allocsz * sizeof *it);
		for (i = 0; i < allocsz; i++) {
			free_item(&it[i]);
		}
		allocsz *= 2;
	}

	/* Pull an item off the free list */
	it = item_free_list;
	item_free_list = it->i_next;
	it->i_next = NULL;

	it->i_prod = p;
	it->i_dot = dot;
	it->i_follow = follow;

	return it;
}

/* Return an item to the free list */
void
free_item(it)
	item_t *it;
{
	it->i_next = item_free_list;
	item_free_list = it;
}

/* Return all items on a list to the free list */
void
free_item_list(its)
	item_t *its;
{
	item_t *nit;

	while (its != NULL) {
	    nit = its->i_next;
	    free_item(its);
	    its = nit;
	}
}
