/*	$Id: lalr_core.c,v 1.2 1999/10/06 00:01:59 leonard Exp $	*/
/* Copyright 1999, David Leonard. The University of Queensland, Australia. */

#include <stdio.h>
#include "Python.h"
#include "structmember.h"	/* for the offsetof() macro */
#include "symbol.h"
#include "lalr.h"
#include "run.h"

/* Identity function. */

PyObject * identity;

PyObject *
identity_func(self, args)
	PyObject *self;	/* unused */
	PyObject *args;	
{
	PyObject *res;

	if (PyTuple_Size(args) == 1)
		res = PyTuple_GetItem(args, 0);
	else
		res = args;

	Py_INCREF(res);
	return res;
}

static char
identity_doc[] =
"identity(arg, ...) -> (arg, ...)\n"
"\tIdentity function returns its single argument, or a tuple";

static struct PyMethodDef
identity_methdef = { "identity", identity_func, METH_VARARGS, identity_doc };

/* Exceptions. */

static PyObject *ConflictErrorObject;
static PyObject *ParseErrorObject;
static PyObject *EmptyTuple;
static PyObject *Zero;
static PyObject *LeftAssoc;
static PyObject *RightAssoc;
static PyObject *NonAssoc;

/* The grammar python object wrapper. */

typedef struct {
	PyObject_HEAD
	PyObject	*go_tlist;		/* terminals */
	PyObject	*go_tname;		/*    their names */
	PyObject	*go_tprec;		/*    their precedence */
	PyObject	*go_tassoc;		/*    their associativity */
	PyObject	*go_ntlist;		/* non-terminals */
	PyObject	*go_ntname;		/*    their names */
	PyObject	*go_prodlist;		/* productions */
	PyObject	*go_eof;		/* end of file terminal */
	PyObject	*go_goal;		/* goal symbol */
	int		go_locked;		/* parser running */
	int		go_nomem;		/* set by nocore() */
	int		go_built;		/* true if rebuild not needed */
	grammar_t	go_grammar;
} GrammarObject;

/* Called by the parser when there is no memory left */
void
nocore(g)
	grammar_t *g;
{
	GrammarObject *go;

	go = (GrammarObject *)((char *)g - offsetof(GrammarObject, go_grammar));
	go->go_nomem = 1;
}

staticforward PyTypeObject GrammarObject_Type;

static void
free_grammar(g)
	grammar_t *g;
{
	state_t *st;

	/* Free all states. */
	while (g->g_state) {
		st = g->g_state->s_next;
		free_state(g->g_state);
		g->g_state = st;
	}
	/* Free the production array. */
	if (g->g_prod) {
		free(g->g_prod);
		g->g_prod = NULL;
	}
	/* Free the name array. */
	if (g->g_name) {
		free(g->g_name);
		g->g_name = NULL;
	}
}

/* Deallocate grammar object. */
static void
G_dealloc(go)
	GrammarObject *go;
{
	grammar_t *g = &go->go_grammar;

	free_grammar(g);

	/* Free the user-provided structures. */
	Py_DECREF(go->go_tlist);
	Py_DECREF(go->go_tname);
	Py_DECREF(go->go_tprec);
	Py_DECREF(go->go_tassoc);
	Py_DECREF(go->go_ntlist);
	Py_DECREF(go->go_ntname);
	Py_DECREF(go->go_prodlist);
	Py_XDECREF(go->go_eof);
	Py_XDECREF(go->go_goal);
	PyMem_DEL(go);
}

/* Return true if the grammar object is in use. */
int
check_locked(go)
	GrammarObject *go;
{
	if (go->go_locked) {
		PyErr_SetString(PyExc_ValueError, "grammar object in use");
		return 1;
	} else
		return 0;
}

/* helper routine for adding terminals and non-terminals */
static PyObject *
add_symbol(go, args, list, namelist, assoclist, preclist)
	GrammarObject *go;
	PyObject *args;
	PyObject *list;
	PyObject *namelist;
	PyObject *assoclist;
	PyObject *preclist;
{
	PyObject *obj;
	PyObject *objname = NULL;
	PyObject *prec = Zero;
	PyObject *assoc = NonAssoc;

	if (assoclist && preclist) {
		if (!PyArg_ParseTuple(args, "O|OOO", &obj, &objname, 
		    &prec, &assoc))
			return NULL;
	} else {
		if (!PyArg_ParseTuple(args, "O|O", &obj, &objname))
			return NULL;
	}

	if (check_locked(go))
		return NULL;

	if (PySequence_In(go->go_tlist, obj) == 1 ||
	    PySequence_In(go->go_ntlist, obj) == 1) {
		PyErr_SetString(PyExc_ValueError,
			"object already in list of symbols");
		return NULL;
	}

	if (assoc != NonAssoc && assoc != LeftAssoc && assoc != RightAssoc) {
		PyErr_SetString(PyExc_TypeError,
			"expected one of NonAssoc, LeftAssoc or RightAssoc "
			"for third argument");
		return NULL;
	}

	prec = PyNumber_Int(prec);
	if (prec == NULL || !PyInt_Check(prec)) {
		PyErr_SetString(PyExc_TypeError,
			"expected integer for fourth argument");
		Py_XDECREF(prec);
		return NULL;
	}

	if (objname == NULL)
		objname = PyObject_Str(obj);
	else
		Py_INCREF(objname);

	if (objname == NULL || !PyString_Check(objname)) {
		PyErr_SetString(PyExc_TypeError, 
		    "expected string for second argument");
		Py_XDECREF(objname);
		Py_DECREF(prec);
		return NULL;
	}

	if (PyList_Size(go->go_tlist) + PyList_Size(go->go_ntlist) >= 
		MAXSYMS - 1) {
		PyErr_SetString(PyExc_OverflowError,
		    "symbol table full");
		Py_DECREF(objname);
		Py_DECREF(prec);
		return NULL;
	}

	Py_INCREF(obj);
	PyList_Append(list, obj);

	PyList_Append(namelist, objname);

	if (preclist && assoclist) {
		Py_INCREF(assoc);
		PyList_Append(preclist, prec);
		PyList_Append(assoclist, assoc);
	} else
		Py_DECREF(prec);

	go->go_built = 0;

	Py_INCREF(Py_None);
	return Py_None;
}

/* Add terminal. */
static PyObject *
G_add_t(go, args)
	GrammarObject *go;
	PyObject *args;
{
	return add_symbol(go, args, go->go_tlist, go->go_tname, 
		go->go_tassoc, go->go_tprec);
}

static char
G_add_t_doc[] =
"add_terminal(object [, string [, int [, assoc]]]) -> None\n"
"\tAdd a terminal symbol, its name, precedence and associativity.";

/* Add non-terminal. */
static PyObject *
G_add_nt(go, args)
	GrammarObject *go;
	PyObject *args;
{
	return add_symbol(go, args, go->go_ntlist, go->go_ntname, 
		NULL, NULL);
}

static char
G_add_nt_doc[] =
"add_nonterminal(object [, string]) -> None\n"
"\tAdd a nonterminal symbol and its name into the grammar.";

/* Add production. */
static PyObject *
G_add_p(go, args)
	GrammarObject *go;
	PyObject *args;
{
	PyObject *lhs, *rhs, *rfn = Py_None, *prec = Py_None;
	int i;
	PyObject *sym;
	PyObject *nrhs;
	PyObject *group;
	int len;

	if (!PyArg_ParseTuple(args, "OO|OO", &lhs, &rhs, &rfn, &prec))
		return NULL;

	/* Check that the lhs is in the terminal or non terminal list */
	if (PySequence_In(go->go_ntlist, lhs) != 1) {
		PyErr_SetObject(PyExc_ValueError,
		    Py_BuildValue("(s,O)",
		        "expected non-terminal for 1st argument", lhs));
		return NULL;
	}

	/* Check the rhs is composed of valid symbols.*/
	if (!PySequence_Check(rhs)) {
		PyErr_SetString(PyExc_TypeError, 
		    "expected sequence for 2nd arg");
		return NULL;
	}

	len = PySequence_Length(rhs);
	if (len > MAXRHS) {
		PyErr_SetString(PyExc_ValueError, "production too long");
		return NULL;
	}

	if (prec == Py_None) 
		Py_INCREF(prec);
	else {
		prec = PyNumber_Int(prec);
		if (prec == NULL || !PyInt_Check(prec)) {
			Py_XDECREF(prec);
			PyErr_Clear();
			PyErr_SetString(PyExc_ValueError, 
			    "expected integer for fourth argument");
			return NULL;
		}
	}

	for (i = 0; i < len; i++) {
		sym = PySequence_GetItem(rhs, i);
		if (PySequence_In(go->go_ntlist, sym) == 1)
			continue;
		if (PySequence_In(go->go_tlist, sym) == 1)
			continue;
		PyErr_SetObject(PyExc_ValueError, 
		     Py_BuildValue("(s,O)",
			 "unknown symbol inside second argument", sym));
		return NULL;
	}

	/* Use the identity function if None is given */
	if (rfn == Py_None) {
		rfn = identity;
		Py_INCREF(identity);
	} 

	/* Make sure that the function is callable */
	if (!PyCallable_Check(rfn)) {
		PyErr_SetObject(PyExc_TypeError,
		     Py_BuildValue("(s,O)",
		        "expected callable object for 3rd arg", rfn));
		return NULL;
	}

	/* Duplicate the rhs */
	nrhs = PyList_New(len);
	for (i = 0; i < len; i++) {
		sym = PySequence_GetItem(rhs, i);
		PyList_SetItem(nrhs, i, sym);
	}

	Py_INCREF(rfn);
	Py_INCREF(lhs);

	/* Construct a handy tuple object. */
	group = PyTuple_New(4);
	PyTuple_SET_ITEM(group, 0, lhs);
	PyTuple_SET_ITEM(group, 1, nrhs);
	PyTuple_SET_ITEM(group, 2, rfn);
	PyTuple_SET_ITEM(group, 3, prec);

	PyList_Append(go->go_prodlist, group);

	go->go_built = 0;

	Py_INCREF(Py_None);
	return Py_None;
}

static char
G_add_p_doc[] =
"add_production(object, list [,func [, prec]]) -> None\n"
"\tAdd a production and an optional evaluation function into the grammar.";

static symbol_t
symtonum(go, sym)
	GrammarObject *go;
	PyObject *sym;
{
	int index;
	int tlen, ntlen;

	if (sym == NULL)
		return (symbol_t)-1;

	tlen = PyList_Size(go->go_tlist);
	for (index = 0; index < tlen; index++)
		if (PyList_GetItem(go->go_tlist, index) == sym) {
			return (symbol_t)(index + 1);
	}

	ntlen = PyList_Size(go->go_ntlist);
	for (index = 0; index < ntlen; index++)
		if (PyList_GetItem(go->go_ntlist, index) == sym) {
			return (symbol_t)(index + tlen + 1);
	}

	PyErr_SetObject(PyExc_ValueError, 
		     Py_BuildValue("(s,O)", "unknown symbol", sym));
	return (symbol_t)-1;
}

/* Called each time for lookahead from the parser. */
static symbol_t
my_lex(arg, valuep)
	void *arg;
	void **valuep;
{
	PyObject *nexttoken = ((PyObject **)arg)[0];
	GrammarObject *go = (GrammarObject*)((PyObject **)arg)[1];
	PyObject **resultp = (PyObject **)valuep;
	PyObject *res;
	symbol_t sym;

	/* call the nexttok */
	res = PyObject_CallObject(nexttoken, EmptyTuple);
	if (res == NULL) {
		*resultp = NULL;
		return -1;
	}
	if (!PyTuple_Check(res) || 
	    (PyTuple_Size(res) != 2 && PyTuple_Size(res) != 1)) {
		PyErr_SetObject(PyExc_ValueError, 
			Py_BuildValue("(s,O,O)",
			    "expected small tuple as return value",
			    nexttoken, res));
		*resultp = NULL;
		return -1;
	}
	sym = symtonum(go, PyTuple_GetItem(res, 0));
	if (PyTuple_Size(res) == 2)
		*resultp = PyTuple_GetItem(res, 1);
	else
		*resultp = Py_None;
	Py_INCREF(*resultp);
	Py_DECREF(res);

	return sym;
}

static void *
my_reduce(p, values)
	production_t *p;
	void **values;
{
	PyObject **rhs = (PyObject **)values;
	PyObject *args, *res;
	PyObject *rfn = (PyObject *)p->p_context;
	int i;

	args = PyTuple_New(p->p_nrhs);
	for (i = 0; i < p->p_nrhs; i++)
	{
		if (rhs[i] == NULL)
			PyTuple_SET_ITEM(args, i, Py_None);
		else
			PyTuple_SET_ITEM(args, i, rhs[i]);
	}
#ifndef NDEBUG
	fprintf(stderr, "calling ");
        PyObject_Print(rfn, stderr, Py_PRINT_RAW);
        PyObject_Print(args, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	if (!PyCallable_Check(rfn)) {
		PyErr_SetString(PyExc_TypeError,
			"expected callable function");
		return NULL;
	}
#endif
	res = PyObject_CallObject(rfn, args);
	Py_DECREF(args);

#ifndef NDEBUG
	fprintf(stderr, "returning ");
        PyObject_Print(res, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
#endif

	return (void *)res;
}

/* Construct parser and begin parsing. */
static PyObject *
G_parse(go, args)
	GrammarObject *go;
	PyObject *args;
{
	PyObject *goal;
	PyObject *nexttoken;
	grammar_t *g = &go->go_grammar;
	production_t *p;
	item_t *it;
	state_t *st;
	PyObject *po;
	PyObject *context[2];
	int numterms, numnonterms, numprods, i, j;
	symbol_t eofsym, sym;
	PyObject *assoc;
	PyObject *result;

	if (!PyArg_ParseTuple(args, "O", &nexttoken))
		return NULL;

	/* Ensure that the nexttoken function is callable */
	if (!PyCallable_Check(nexttoken)) {
		PyErr_SetString(PyExc_TypeError,
			"expected callable function for second argument");
		return NULL;
	}

	/* Recursive parse? */
	if (check_locked(go))
		return NULL;

	if (go->go_eof == NULL) {
		PyErr_SetString(PyExc_RuntimeError,
			"EOF symbol has not been set with set_eof()");
		return NULL;
	}

	/* Okay, there's no turning back now.... */

	go->go_locked = 1;

	if (!go->go_built) {
		free_grammar(g);

		numterms = PySequence_Length(go->go_tlist);
		numnonterms = PySequence_Length(go->go_ntlist);
		numprods = PySequence_Length(go->go_prodlist);

		g->g_terminals = 1 + numterms;
		g->g_nonterminals = 1 + numterms + numnonterms;
		if ((g->g_name = Py_Malloc(g->g_nonterminals * sizeof(char *)))
		    == NULL) goto error;
		if ((g->g_prod = Py_Malloc(numprods * sizeof(production_t)))
		    == NULL) goto error;
		g->g_laststate = NULL;
		g->g_rrconflicts = 0;
		g->g_srconflicts = 0;

		/* Set up the diagnostic names */
		g->g_name[0] = "<epsilon>";
		for (i = 0; i < numterms; i++) {
			g->g_name[1 + i] = 
			    PyString_AsString(PyList_GetItem(go->go_tname, i));
			g->g_symprec[1 + i] =
			    PyInt_AsLong(PyList_GetItem(go->go_tprec, i));
			assoc = PyList_GetItem(go->go_tassoc, i);
			g->g_symassoc[1 + i] =
				(assoc == LeftAssoc)  ? LEFTASSOC :
				(assoc == RightAssoc) ? RIGHTASSOC : 
						        NONASSOC;
		}
		for (i = 0; i < numnonterms; i++)
			g->g_name[1 + numterms + i] = 
			    PyString_AsString(PyList_GetItem(go->go_ntname, i));

		/* Set up the productions */
		for (i = 0; i < numprods; i++) {
			PyObject *lhs;
			PyObject *rhs;
			PyObject *rfn;
			PyObject *prec;

			p = &g->g_prod[i];

			po = PySequence_GetItem(go->go_prodlist, i);
			lhs = PyTuple_GetItem(po, 0);
			rhs = PyTuple_GetItem(po, 1);
			rfn = PyTuple_GetItem(po, 2);
			prec = PyTuple_GetItem(po, 3);

			p->p_context = (void *)rfn;
			p->p_next_prod = &g->g_prod[i+1];

			/* Fill in the symbols. */
			p->p_lhs = symtonum(go, lhs);
			if (p->p_lhs == -1) 
				goto error;
			p->p_nrhs = PyList_Size(rhs);
			for (j = 0; j < p->p_nrhs; j++) {
				p->p_rhs[j] = symtonum(go, 
				    PyList_GetItem(rhs, j));
				if (p->p_rhs[j] == -1)
					goto error;
			}

			/* Compute this production's precedence */
			if (prec == Py_None) {
				/*
				 * Use the precedence level of the
				 * last terminal (or 0 if no terminals)
				 */
				p->p_prec = 0;
				for (j = p->p_nrhs - 1; j >= 0; j--) 
				    if (p->p_rhs[j] < g->g_terminals) {
					p->p_prec = g->g_symprec[p->p_rhs[j]];
					break;
				    }
			} else
				p->p_prec = PyInt_AsLong(prec);
		}
		g->g_prod[numprods-1].p_next_prod = NULL;

		/* Derive productions */
		derive_productions(g);

		/* Derive firsts */
		derive_firsts(g);

		/* Create initial state */
		it = NULL;
		eofsym = symtonum(go, go->go_eof);
		for (p = g->g_derive[symtonum(go, go->go_goal)]; p;
		    p = p->p_next_derive) {
			item_t *nit;
			nit = new_item(p, 0, eofsym);
			nit->i_next = it;
			it = nit;
		}
		st = new_state(g, it);
		g->g_init = g->g_newstates = st;
		st->s_next_new = NULL;

		go->go_nomem = 0;

		/* form other states */
		while ((st = g->g_newstates) != NULL) {
			g->g_newstates = st->s_next_new;
			st->s_next_new = st;
			closure(g, st);

			add_transitions(g, st);
			if (go->go_nomem) {
				go->go_nomem = 0;
				goto error;
			}
		}

#ifndef NDEBUG
		for (st = g->g_state; st; st = st->s_next)
			print_state(g, st);
		fprintf(stderr, "Conflicts: %d shift-reduce, "
			"%d reduce-reduce\n",
			g->g_srconflicts, g->g_rrconflicts);
#endif
		go->go_built = 1;
	}

	/* perform the actual parse */
	context[0] = nexttoken;
	context[1] = (PyObject *)go;
	switch (run(g, my_lex, &context, my_reduce, (void **)&result)) {
	case RUN_ACCEPT:
		break;
	case RUN_SYNTAX:
		PyErr_SetObject(ParseErrorObject, result);
		goto error;
	case RUN_INTERNAL:
		goto error;
#ifndef NDEBUG
	default:
		abort();
#endif
	}

	go->go_locked = 0;
	return result;

error:
	go->go_locked = 0;
	return NULL;
}

static char
G_parse_doc[] =
"parse(func) -> result\n"
"\tPerform a parse, obtaining tokens tuples from the function.";

/* Set the eof symbol. */
static PyObject *
G_goal_eof(go, args)
	GrammarObject *go;
	PyObject *args;
{
	PyObject *goal;
	PyObject *eof;

	if (!PyArg_ParseTuple(args, "OO", &goal, &eof))
		return NULL;

	if (check_locked(go))
		return NULL;

	if (PySequence_In(go->go_ntlist, goal) != 1) {
		PyErr_SetString(PyExc_ValueError, 
			"expected a known nonterminal for the goal");
		return NULL;
	}

	if (PySequence_In(go->go_tlist, eof) != 1) {
		PyErr_SetString(PyExc_ValueError, 
			"expected a known terminal for eof");
		return NULL;
	}

	Py_XDECREF(go->go_eof);
	Py_XDECREF(go->go_goal);
	Py_INCREF(eof);
	Py_INCREF(goal);
	go->go_eof = eof;
	go->go_goal = goal;

	Py_INCREF(Py_None);
	return Py_None;
}

static char
G_goal_eof_doc[] =
"set_goal_eof(object, object) -> None\n"
"\tSet the goal symbol of the grammar and the terminal expected to follow it.";

static PyMethodDef GrammarObject_methods[] = {
	{ "add_terminal",	G_add_t,  METH_VARARGS, G_add_t_doc },
	{ "add_nonterminal",	G_add_nt, METH_VARARGS, G_add_nt_doc },
	{ "add_production",	G_add_p,  METH_VARARGS, G_add_p_doc },
	{ "parse",		G_parse,  METH_VARARGS, G_parse_doc },
	{ "set_goal_eof",	G_goal_eof, METH_VARARGS, G_goal_eof_doc },
	{ NULL, NULL }
};

static PyObject *
G_getattr(go, name)
	GrammarObject *go;
	char *name;
{
	return Py_FindMethod(GrammarObject_methods, (PyObject *)go, name);
}

static PyObject *
G_setattr(go, name, v)
	GrammarObject *go;
	char *name;
	PyObject *v;
{
	PyErr_SetNone(PyExc_AttributeError);
	return NULL;
}

statichere PyTypeObject 
GrammarObject_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"LALR(1) Parser",	/*tp_name*/
	sizeof(GrammarObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)G_dealloc,  /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)G_getattr,	/*tp_getattr*/
	(setattrfunc)G_setattr,	/*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};

static PyObject *
new_parser(self, args)
	PyObject *self; /* unused */
	PyObject *args;
{
	GrammarObject *go;
	grammar_t *g;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	go = PyObject_NEW(GrammarObject, &GrammarObject_Type);
	if (go == NULL)
		return NULL;

	go->go_tlist = PyList_New(0);
	go->go_tname = PyList_New(0);
	go->go_tprec = PyList_New(0);
	go->go_tassoc = PyList_New(0);
	go->go_ntlist = PyList_New(0);
	go->go_ntname = PyList_New(0);
	go->go_prodlist = PyList_New(0);
	go->go_locked = 0;
	go->go_goal = NULL;
	go->go_eof = NULL;
	go->go_built = 0;

	g = &go->go_grammar;
	g->g_state = NULL;
	g->g_prod = NULL;
	g->g_name = NULL;

	return (PyObject *)go;
}
static char
 new_parser_doc[] =
"new_parser() -> object\n"
"\tCreate a new LALR(1) grammar object.";

/* Functions defined in this module */
static PyMethodDef
 module_methods[] = {
	{ "new_parser",	new_parser,	METH_VARARGS, new_parser_doc },
	{ NULL, NULL }
};

DL_EXPORT(void)
initlalr_core()
{
	PyObject *m, *d;

	GrammarObject_Type.ob_type = &PyType_Type;
	m = Py_InitModule("lalr_core", module_methods);
	d = PyModule_GetDict(m);

	/* Add the two error objects */
	ConflictErrorObject = PyErr_NewException("lalr_core.ConflictError", 
		NULL, NULL);
	ParseErrorObject = PyErr_NewException("lalr_core.ParseError", 
		NULL, NULL);

	PyDict_SetItemString(d, "ConflictError", ConflictErrorObject);
	PyDict_SetItemString(d, "ParseError", ParseErrorObject);

	/* Add some extra helpful values */
	LeftAssoc = PyString_FromString("LeftAssoc");
	RightAssoc = PyString_FromString("RightAssoc");;
	NonAssoc = PyString_FromString("NonAssoc");;

	PyDict_SetItemString(d, "LeftAssoc", LeftAssoc);
	PyDict_SetItemString(d, "RightAssoc", RightAssoc);
	PyDict_SetItemString(d, "NonAssoc", NonAssoc);

	/* A fast identity function. */
	identity = PyCFunction_New(&identity_methdef, NULL);

	EmptyTuple = PyTuple_New(0);
	Zero = PyInt_FromLong(0);

}
