/*	$Id: pyrun.c,v 1.1.1.1 1999/10/05 23:58:37 leonard Exp $	*/
/* Copyright 1999, David Leonard. The University of Queensland, Australia. */

#include <stdio.h>
#include <assert.h>
#include "Python.h"
#include "symbol.h"
#include "lalr.h"
#include "run.h"

#define MAXSTACK 256

/*
 * Run the parser state machine. A stack of symbol/state pairs is
 * created and shifts and reduce-shifts are efficiently called.
 * We call the input function (ifn) to get new symbols when needed.
 */
int
run(g, ifn, istate, redfn, resultp)
	grammar_t *g;
	symbol_t (* ifn)(void *, void **value);
	void *istate;
	void* redfn(production_t *, void **value);
	void **resultp;
{
	PyObject *expect;
	symbol_t next, sym;
	void *   next_value;
	void *   rval;
	symbol_t sym_stack[MAXSTACK];
	state_t  *state_stack[MAXSTACK];
	void	 *value_stack[MAXSTACK];
	state_t  *st;
	production_t *p;
	int top;
	int retval;
	int i;

	top = 0;

	state_stack[top] = g->g_init;
	value_stack[top] = NULL;
	sym_stack[top] = -1;

	next = (*ifn)(istate, &next_value);
	if (next == -1) {
#ifndef NDEBUG
		fprintf(stderr, "exception from first token\n");
#endif
		retval = RUN_INTERNAL;
		goto error;
	}

	for (;;) {
#ifndef NDEBUG
		for (i = 0; i <= top; i++) {
			fprintf(stderr, " %s/%d", 
				(i)?g->g_name[sym_stack[i]]:"",
				state_stack[i] ? state_stack[i]->s_id : -1);
			fprintf(stderr, "/");
			PyObject_Print((PyObject *)value_stack[i], stderr,
				Py_PRINT_RAW);
		}
		fprintf(stderr, " | %s\n", g->g_name[next]);
#endif
		st = state_stack[top]->s_shift[next];
		p = state_stack[top]->s_reduce[next];

		if (st != NULL) {
			/* shift input */
			top++;
			state_stack[top] = st;
			sym_stack[top] = next;
			value_stack[top] = next_value;
#ifndef NDEBUG
			fprintf(stderr, "shift %s/", g->g_name[next]);
			PyObject_Print((PyObject *)next_value, stderr,
				Py_PRINT_RAW);
			fprintf(stderr, "\n");
#endif
			next = (*ifn)(istate, &next_value);
			if (next == -1) {
				/* token generator error */
				retval = RUN_INTERNAL;
				goto error;
			}

		} else if (p != NULL) {
			/* reduce to nt */
#ifndef NDEBUG
			fprintf(stderr, "reduce ");
			print_production(g, p);
			fprintf(stderr, "\n");
#endif
			top -= p->p_nrhs;
			if (redfn)
				rval = (*redfn)(p, &value_stack[top + 1]);
			else {
				Py_INCREF(Py_None);
				rval = Py_None;
			}

			if (rval == NULL) {
				retval = RUN_INTERNAL;
				goto error;
			}

			/* shift nt */
			st = state_stack[top]->s_shift[p->p_lhs];
			if (top == 0 && st == NULL) {
				/* shifted goal symbol */
#ifndef NDEBUG
				fprintf(stderr, "accept\n");
#endif
				if (resultp)
					*resultp = rval;
				return RUN_ACCEPT;
			}
#ifndef NDEBUG
			if (st == NULL) {
				PyErr_SetString(PyExc_RuntimeError,
					"parser broken");
				retval = RUN_INTERNAL;
				goto error;
			}
#else
			assert(st);
#endif
			top++;
			state_stack[top] = st;
			value_stack[top] = rval;
			sym_stack[top] = p->p_lhs;
#ifndef NDEBUG
			fprintf(stderr, "shift %s/", g->g_name[p->p_lhs]);
			PyObject_Print((PyObject *)rval, stderr,
				Py_PRINT_RAW);
			fprintf(stderr, "\n");
#endif
		} else {
			/* not in lookahead set! */
#ifndef NDEBUG
			fprintf(stderr, "syntax error on %s\n", g->g_name[next]);
#endif
			/*
			 * POSSIBLE IDEAS:
			 * if there is an error symbol shiftable in this state,
			 * fake and 'shift' it. when in the new state, 
			 * consume symbols until something comes up in either
			 * a shift or reduce set?
			 */
			/*
			 * For now, crash and burn. Later, will need to
			 * call some error handler that can know more about
			 * the symbol (maybe it's file position?)
			 */
			/* XXXXXXXX memory leak */
			if (next == RUN_ACCEPT) {
				retval = RUN_INTERNAL;
			}

			expect = PyList_New(0);
			for (sym = 1; sym < g->g_terminals; sym++)
			    if (state_stack[top]->s_shift[sym])
				PyList_Append(expect, 
				    Py_BuildValue("s", g->g_name[sym]));
			st = state_stack[top]->s_shift[next];
			p = state_stack[top]->s_reduce[next];
			*resultp = Py_BuildValue("OOs", expect, next_value, 
				g->g_name[next]);
			retval = RUN_SYNTAX;
			goto error;
		}
	}

error:
	/* Clean up python values */
	for (i = 0; i <= top; i++)
		Py_XDECREF((PyObject *)value_stack[i]);
	return retval;
}
